import sys
from pprint import pformat

from pyramid.tweens import EXCVIEW
from pyramid.settings import aslist
from pyramid.settings import asbool
from pyramid.util import DottedNameResolver
from pyramid.httpexceptions import WSGIHTTPException
from pyramid.security import unauthenticated_userid

resolver = DottedNameResolver(None)

PY3 = sys.version_info[0] == 3

if PY3: # pragma: no cover
    import builtins
else:
    import __builtin__ as builtins

def as_globals_list(value):
    L = []
    value = aslist(value)
    for dottedname in value:
        if dottedname in builtins.__dict__:
            if PY3: # pragma: no cover
                dottedname = 'builtins.%s' % dottedname
            else:
                dottedname = '__builtin__.%s' % dottedname
        obj = resolver.resolve(dottedname)
        L.append(obj)
    return L

def _get_url(request):
    try:
        url = request.url
    except UnicodeDecodeError:
        # do the best we can
        url = (request.host_url +
               request.environ.get('SCRIPT_NAME') +
               request.environ.get('PATH_INFO'))
        qs = request.environ.get('QUERY_STRING')
        if qs:
            url += '?' + qs
        url = 'could not decode url: %r' % url
    return url

_MESSAGE_TEMPLATE = """

%(url)s

ENVIRONMENT

%(env)s


PARAMETERS

%(params)s


UNAUTHENTICATED USER

%(usr)s

"""

def _get_message(request):
    url = _get_url(request)
    unauth = unauthenticated_userid(request)

    try:
        params = request.params
    except UnicodeDecodeError:
        params = 'could not decode params'

    return _MESSAGE_TEMPLATE % dict(
            url=url,
            env=pformat(request.environ),
            params=pformat(params),
            usr=unauth if unauth else '')

def _handle_error(request, getLogger, get_message):
    # save the traceback as it may get lost when we get the message.
    # _handle_error is not in the traceback, so calling sys.exc_info
    # does NOT create a circular reference
    exc_info = sys.exc_info()
    try:
        logger = getLogger('exc_logger')
        message = get_message(request)
        logger.error(message, exc_info=exc_info)
    except:
        logger.exception("Exception while logging")
    raise

def exclog_tween_factory(handler, registry):

    get = registry.settings.get

    ignored = get('exclog.ignore', (WSGIHTTPException,))
    get_message = _get_url
    if get('exclog.extra_info', False):
         get_message = _get_message
    get_message = get('exclog.get_message', get_message)

    getLogger = get('exclog.getLogger', 'logging.getLogger')
    getLogger = resolver.resolve(getLogger)

    def exclog_tween(request, getLogger=getLogger):
        # getLogger injected for testing purposes
        try:
            return handler(request)
        except ignored:
            raise
        except:
            _handle_error(request, getLogger, get_message)
            # _handle_error always raises
            raise AssertionError('Should never get here') #pragma NO COVER
    return exclog_tween

def includeme(config):
    """
    Set up am implicit :term:`tween` to log exception information that is
    generated by your Pyramid application.  The logging data will be sent to
    the Python logger named ``exc_logger``.

    This tween configured to be placed 'below' the exception view tween.  It
    will log all exceptions (even those eventually caught by a Pyramid
    exception view) except 'http exceptions' (any exception that derives from
    ``pyramid.httpexceptions.WSGIHTTPException`` such as ``HTTPFound``).  You
    can instruct ``pyramid_exclog`` to ignore custom exception types by using
    the ``excview.ignore`` configuration setting.
    """
    get = config.registry.settings.get
    ignored = as_globals_list(get('exclog.ignore',
                                  'pyramid.httpexceptions.WSGIHTTPException'))
    extra_info = asbool(get('exclog.extra_info', False))
    get_message = get('exclog.get_message', None)
    if get_message is not None:
        get_message = config.maybe_dotted(get_message)
        config.registry.settings['exclog.get_message'] = get_message
    config.registry.settings['exclog.ignore'] = tuple(ignored)
    config.registry.settings['exclog.extra_info'] = extra_info
    config.add_tween('pyramid_exclog.exclog_tween_factory', under=EXCVIEW)



import sys
import unittest
from pyramid import testing

class Test_exclog_tween_factory(unittest.TestCase):
    def setUp(self):
        self.config = testing.setUp()
        self.registry = self.config.registry

    def _callFUT(self):
        from pyramid_exclog import exclog_tween_factory
        return exclog_tween_factory(None, self.registry)

    def test_no_recipients(self):
        handler = self._callFUT()
        self.assertEqual(handler.__name__, 'exclog_tween')


class Test_exclog_tween(unittest.TestCase):
    def setUp(self):
        from pyramid.request import Request
        request = Request.blank('/')
        self.request = request
        self.config = testing.setUp(request=request)
        self.registry = self.config.registry
        self.registry.settings = {}
        request.registry = self.registry
        self.logger = DummyLogger()

    def getLogger(self, name):
        self.logger.name = name
        return self.logger

    def handler(self, request):
        raise NotImplementedError

    def _callFUT(self, handler=None, registry=None, request=None,
                 getLogger=None):
        from pyramid_exclog import exclog_tween_factory
        if handler is None:
            handler = self.handler
        if registry is None:
            registry = self.registry
        if request is None:
            request = self.request
        if getLogger is None:
            getLogger = self.getLogger
        tween = exclog_tween_factory(handler, registry)
        return tween(request, getLogger=getLogger)

    def test_ignored(self):
        self.registry.settings['exclog.ignore'] = (NotImplementedError,)
        self.assertRaises(NotImplementedError, self._callFUT)
        self.assertEqual(len(self.logger.exceptions), 0)

    def test_notignored(self):
        self.assertRaises(NotImplementedError, self._callFUT)
        self.assertEqual(len(self.logger.exceptions), 1)
        msg = self.logger.exceptions[0]
        self.assertEqual(msg, self.request.url)

    def test_extra_info(self):
        self.registry.settings['exclog.extra_info'] = True
        self.assertRaises(NotImplementedError, self._callFUT)
        self.assertEqual(len(self.logger.exceptions), 1)
        msg = self.logger.exceptions[0].strip()
        self.assertTrue(msg.startswith("http://localhost/\n\nENVIRONMENT"))
        self.assertTrue("PARAMETERS\n\nNestedMultiDict([])" in msg)
        self.assertTrue('ENVIRONMENT' in msg)

    def test_get_message(self):
        self.registry.settings['exclog.get_message'] = lambda req: 'MESSAGE'
        self.assertRaises(NotImplementedError, self._callFUT)
        self.assertEqual(len(self.logger.exceptions), 1)
        msg = self.logger.exceptions[0]
        self.assertEqual(msg, 'MESSAGE')

    def test_user_info_user(self):
        self.config.testing_securitypolicy(
                userid='hank',
                permissive=True)
        self.registry.settings['exclog.extra_info'] = True
        self.assertRaises(NotImplementedError, self._callFUT)
        self.assertEqual(len(self.logger.exceptions), 1)
        msg = self.logger.exceptions[0]
        self.assertTrue('UNAUTHENTICATED USER\n\nhank' in msg)

    def test_user_info_no_user(self):
        self.registry.settings['exclog.extra_info'] = True
        self.assertRaises(NotImplementedError, self._callFUT)
        self.assertEqual(len(self.logger.exceptions), 1)
        msg = self.logger.exceptions[0]
        self.assertTrue('UNAUTHENTICATED USER\n\n\n' in msg)

    def test_exception_while_logging(self):
        from pyramid.request import Request
        bang = AssertionError('bang')
        class BadRequest(Request):
            @property
            def url(self):
                raise bang
        request = BadRequest.blank('/')
        self.assertRaises(Exception, self._callFUT, request=request)
        msg = self.logger.exceptions[0]
        self.assertEqual('Exception while logging', msg)
        raised = self.logger.exc_info[0][1]
        self.assertEqual(raised, bang)


class Test__get_url(unittest.TestCase):

    def _callFUT(self, request):
        from pyramid_exclog import _get_url
        return _get_url(request)

    def test_normal(self):
        from pyramid.testing import DummyRequest
        request = DummyRequest()
        self.assertEqual(self._callFUT(request), 'http://example.com')

    def test_w_deocode_error_wo_qs(self):
        from pyramid.request import Request
        request = Request.blank('/')
        request.environ['SCRIPT_NAME'] = '/script'
        request.environ['PATH_INFO'] = '/path/with/latin1/\x80'
        self.assertEqual(self._callFUT(request),
                         r"could not decode url: " +
                         r"'http://localhost/script/path/with/latin1/\x80'")

    def test_w_deocode_error_w_qs(self):
        from pyramid.request import Request
        request = Request.blank('/')
        request.environ['SCRIPT_NAME'] = '/script'
        request.environ['PATH_INFO'] = '/path/with/latin1/\x80'
        request.environ['QUERY_STRING'] = 'foo=bar'
        self.assertEqual(self._callFUT(request),
                         r"could not decode url: " +
                         r"'http://localhost/script/path/with/latin1/\x80" +
                         r"?foo=bar'")


class Test__get_message(unittest.TestCase):

    def _callFUT(self, request):
        from pyramid_exclog import _get_message
        return _get_message(request)

    def test_evil_encodings(self):
        from pyramid.request import Request
        request = Request.blank('/%FA') # not utf-8
        msg = self._callFUT(request)
        self.assertTrue("could not decode url: 'http://localhost/" in msg)

    def test_evil_encodings_extra_info(self):
        from pyramid.request import Request
        request = Request.blank('/url?%FA=%FA') # not utf-8
        msg = self._callFUT(request)
        self.assertTrue("could not decode params" in msg, msg)

    def test_evil_encodings_extra_info_POST(self):
        from pyramid.request import Request
        request = Request.blank('/url',
                                content_type=
                                    'application/x-www-form-urlencoded; '
                                    'charset=utf-8',
                                POST='%FA=%FA') # not utf-8
        self._callFUT(request) # doesn't fail


class Test_includeme(unittest.TestCase):
    def _callFUT(self, config):
        from pyramid_exclog import includeme
        return includeme(config)

    def test_it(self):
        from pyramid.httpexceptions import WSGIHTTPException
        from pyramid.tweens import EXCVIEW
        config = DummyConfig()
        self._callFUT(config)
        self.assertEqual(config.tweens,
             [('pyramid_exclog.exclog_tween_factory', EXCVIEW, None)])
        self.assertEqual(config.registry.settings['exclog.ignore'],
                         (WSGIHTTPException,))

    def test_it_catchall(self):
        from pyramid.tweens import EXCVIEW
        config = DummyConfig()
        config.settings['exclog.catchall'] = 'true'
        self._callFUT(config)
        self.assertEqual(config.tweens,
                   [('pyramid_exclog.exclog_tween_factory', EXCVIEW,
                     None)])

    def test_it_withignored_builtin(self):
        config = DummyConfig()
        config.settings['exclog.ignore'] = 'NotImplementedError'
        self._callFUT(config)
        self.assertEqual(config.registry.settings['exclog.ignore'],
                         (NotImplementedError,))

    def test_it_withignored_nonbuiltin(self):
        config = DummyConfig()
        config.settings['exclog.ignore'] ='pyramid_exclog.tests.DummyException'
        self._callFUT(config)
        self.assertEqual(config.registry.settings['exclog.ignore'],
                         (DummyException,))

    def test_it_with_extra_info(self):
        config = DummyConfig()
        config.settings['exclog.extra_info'] = 'true'
        self._callFUT(config)
        self.assertEqual(config.registry.settings['exclog.extra_info'], True)

    def test_it_with_get_message(self):
        config = DummyConfig()
        get_message = lambda req: 'MESSAGE'
        config.settings['exclog.get_message'] = get_message
        self._callFUT(config)
        self.assertEqual(config.registry.settings['exclog.get_message'],
                         get_message)

    def test_get_message_not_set_by_includeme(self):
        config = DummyConfig()
        self._callFUT(config)
        self.assertTrue('exclog.get_message' not in config.registry.settings)


class DummyException(object):
    pass

class DummyLogger(object):
    def __init__(self):
        self.exceptions = []
        self.exc_info = []

    def error(self, msg, exc_info=None):
        self.exceptions.append(msg)
        self.exc_info.append(exc_info)

    def exception(self, msg):
        self.exceptions.append(msg)
        self.exc_info.append(sys.exc_info())

class DummyConfig(object):
    def __init__(self):
        self.tweens = []
        self.registry = self
        self.settings = {}

    def add_tween(self, factory, under=None, over=None):
        self.tweens.append((factory, under, over))

    def maybe_dotted(self, obj):
        """NOTE: ``obj`` should NOT be a dotted name."""
        return obj
